/*
 * To change this template, choose Tools | Templates
 * and open the template in the editor.
 *
 * TO IMPLMENT :
 * http://www.cgal.org/Manual/3.3/doc_html/cgal_manual/packages.html#Part:TriangulationsAndDelaunayTriangulations
 * http://www.cs.berkeley.edu/~jrs/meshf99/proj1.html
 Three fixes to Guibas and Stolfi's "Primitives for the Manipulation of
 General Subdivisions and the Computation of Voronoi Diagrams":

 One, their Locate primitive always works in the interior of the mesh,
 but can get stuck in an infinite loop on the convex hull.  Consider
 this mesh:
 y       z
 o-------o
 / \     / \
 /   \   /   \
 /     \ /     \
 o--X----o-------o
 u       v       w

 Suppose you're trying to locate the X, which is collinear with uvw.
 Look at p. 121 of Guibas & Stolfi.  You are searching with an edge
 e for the edge uv.  Suppose that e := wv (directed toward v).
 RightOf[X, e] is false.  NOT RightOf[X, e.Onext] is true, so we
 execute "e <- e.Onext", which sets e to wz.  On the next iteration,
 we execute "e <- e.Onext", which sets e back to wv.  This continues
 forever, an infinite loop.

 To fix this bug, changing both "Not RightOf" expressions to "LeftOf"
 expressions is a good start, but doesn't fulfill the criterion that
 "Locate returns an edge e of the current Delaunay diagram such that
 the given point X is either on e or _strictly_ inside the left face
 of e."  Rather, X might be on any edge of the left face; X might
 even be the apex vertex.  An explicit test for these possibilities
 is necessary.

 Two, I think there are other places in the paper where Guibas &
 Stolfi sometime use RightOf when they mean NOT LeftOf, and vice
 versa.  Give careful consideration in each case to what should
 happen when the determinant is zero.

 Three, on p. 120, in their InsertSite procedure, the fifth-last line
 is wrong.  The statement "e <- t" should read "e <- e.Oprev".

 If you're not using exact arithmetic for the CCW and InCircle tests,
 other things can go wrong as well.  Some problem cases can be
 avoided by changing the code a bit, and some can't.

 */
package com.jdcm.charts;

import java.util.ArrayList;
import java.util.List;

/**
 * Incremental Delaunay Triangulatiuon
 *
 * @author Java-code by X.Philippeau - Pseudo-code by Guibas and Stolfi
 *
 * @see Primitives for the Manipulation of General Subdivisions and the
 * Computation of Voronoi Diagrams (Leonidas Guibas,Jorge Stolfi)
 */
public class A_Delaunay {
    private final BoundingBox bbox = new BoundingBox();
        private final List<QuadEdge> quadEdge = new ArrayList<>();
    // starting edge for walk (see locate() method)
    private QuadEdge startingEdge = null;
    // list of quadEdge belonging to Delaunay triangulation


    // Bounding box of the triangulation
    class BoundingBox {

        double minx, miny, maxx, maxy;
        ValuedPoint2D a = new ValuedPoint2D(); // lower left
        ValuedPoint2D b = new ValuedPoint2D(); // lower right
        ValuedPoint2D c = new ValuedPoint2D(); // upper right
        ValuedPoint2D d = new ValuedPoint2D(); // upper left
    }


    /**
     * Constuctor:
     */
    public A_Delaunay() {

        bbox.minx = Double.MAX_VALUE;
        bbox.maxx = -Double.MAX_VALUE;
        bbox.miny = Double.MAX_VALUE;
        bbox.maxy = -Double.MAX_VALUE;

        // create the QuadEdge graph of the bounding box
        QuadEdge ab = QuadEdge.makeEdge(bbox.a, bbox.b);
        QuadEdge bc = QuadEdge.makeEdge(bbox.b, bbox.c);
        QuadEdge cd = QuadEdge.makeEdge(bbox.c, bbox.d);
        QuadEdge da = QuadEdge.makeEdge(bbox.d, bbox.a);
        QuadEdge.splice(ab.sym(), bc);
        QuadEdge.splice(bc.sym(), cd);
        QuadEdge.splice(cd.sym(), da);
        QuadEdge.splice(da.sym(), ab);

        this.startingEdge = ab;
    }

    /**
     * update the dimension of the bounding box
     *
     * @param minx,miny,maxx,maxy summits of the rectangle
     * @param miny
     * @param maxx
     * @param maxy
     * 
     */
    public void setBoundigBox(double minx, double miny, double maxx, double maxy) {
        // update saved values
        bbox.minx = minx;
        bbox.maxx = maxx;
        bbox.miny = miny;
        bbox.maxy = maxy;

        // extend the bounding-box to surround min/max
        double centerx = (minx + maxx) / 2;
        double centery = (miny + maxy) / 2;
        double x_min = (minx - centerx - 1) * 10 + centerx;
        double x_max = (maxx - centerx + 1) * 10 + centerx;
        double y_min = (miny - centery - 1) * 10 + centery;
        double y_max = (maxy - centery + 1) * 10 + centery;

        // set new positions
        bbox.a.setLocation(x_min, y_min);
        bbox.b.setLocation(x_max, y_min);
        bbox.c.setLocation(x_max, y_max);
        bbox.d.setLocation(x_min, y_max);
    }

    // update the size of the bounding box (cf locate() method)
    private void updateBoundigBox(ValuedPoint2D p) {
        double minx = Math.min(bbox.minx, p.getX());
        double maxx = Math.max(bbox.maxx, p.getX());
        double miny = Math.min(bbox.miny, p.getY());
        double maxy = Math.max(bbox.maxy, p.getY());
        setBoundigBox(minx, miny, maxx, maxy);
        //System.out.println("resizing bounding-box: "+minx+" "+miny+" "+maxx+" "+maxy);
    }

    /**
     * Returns an edge e of the triangle containing the point p (Guibas and
     * Stolfi)
     *
     * @param p the point to localte
     * @return the edge of the triangle
     */
    private QuadEdge locate(ValuedPoint2D p) {

        /* outside the bounding box ? */
        if (p.getX() < bbox.minx || p.getX() > bbox.maxx || p.getY() < bbox.miny || p.getY() > bbox.maxy) {
            updateBoundigBox(p);
        }
        QuadEdge e = startingEdge;
        int i = 0;
        while (true) {
            /* duplicate point ? */
            /* !!! infinite loop here!!!*/

            if (p.getX() == e.orig().getX() && p.getY() == e.orig().getY()) {
                return e;
            }
            if (p.getX() == e.dest().getX() && p.getY() == e.dest().getY()) {
                return e;
            }

            /* walk */
            if (QuadEdge.isAtRightOf(e, p)) {
                e = e.sym();
            } else if (!QuadEdge.isAtRightOf(e.onext(), p)) {
                e = e.onext();
            } else if (!QuadEdge.isAtRightOf(e.dprev(), p)) {
                e = e.dprev();
            } else {
                return e;
            }
            if (i++ > quadEdge.size()) {
                if (QuadEdge.isAtRightOf(e, p)) {
                    e = e.sym();
                    //else if (!QuadEdge.isAtRightOf(e.onext(),p))
                    //	e = e.onext();
                } else if (!QuadEdge.isAtRightOf(e.dprev(), p)) {
                    e = e.dprev();
                } else {
                    return e;
                }
            }
        }
    }

    /**
     * Inserts a new point into a Delaunay triangulation (Guibas and Stolfi)
     *
     * @param p the point to insert
     */
    public void insertPoint(ValuedPoint2D p) {
        //  System.out.println("==>"+p.getX()+","+p.getY()+","+p.getvalue());
        QuadEdge e = locate(p);
        //  System.out.println("===>"+p.getX()+","+p.getY()+","+p.getvalue());
        // point is a duplicate -> nothing to do
        if (p.getX() == e.orig().getX() && p.getY() == e.orig().getY()) {
            return;
        }
        if (p.getX() == e.dest().getX() && p.getY() == e.dest().getY()) {
            return;
        }

        // point is on an existing edge -> remove the edge
        if (QuadEdge.isOnEdge(e, p)) {
            e = e.oprev();
            this.quadEdge.remove(e.onext());
            QuadEdge.deleteEdge(e.onext());
        }

        // Connect the new point to the vertices of the containing triangle
        // (or quadrilateral in case of the point is on an existing edge)
        QuadEdge base = QuadEdge.makeEdge(e.orig(), p);
        this.quadEdge.add(base);

        QuadEdge.splice(base, e);
        this.startingEdge = base;
        do {
            base = QuadEdge.connect(e, base.sym());
            this.quadEdge.add(base);
            e = base.oprev();
        } while (e.lnext() != startingEdge);
        // Examine suspect edges to ensure that the Delaunay condition is satisfied.
        do {
            QuadEdge t = e.oprev();

            if (QuadEdge.isAtRightOf(e, t.dest())
                    && QuadEdge.inCircle(e.orig(), t.dest(), e.dest(), p)) {
                // flip triangles
                QuadEdge.swapEdge(e);
                e = e.oprev();
            } else if (e.onext() == startingEdge) {
                return; // no more suspect edges
            } else {
                e = e.onext().lprev();  // next suspect edge
            }
        } while (true);

    }

    /**
     * compute and return the list of edges
     * 
     * @return 
     */
    public List<ValuedPoint2D[]> computeEdges() {
        List<ValuedPoint2D[]> edges = new ArrayList<>();
        // do not return edges pointing to/from surrouding triangle
        for (QuadEdge q : this.quadEdge) {
            if (q.orig() == bbox.a || q.orig() == bbox.b || q.orig() == bbox.c || q.orig() == bbox.d) {
                continue;
            }
            if (q.dest() == bbox.a || q.dest() == bbox.b || q.dest() == bbox.c || q.dest() == bbox.d) {
                continue;
            }
            edges.add(new ValuedPoint2D[]{q.orig(), q.dest()});
        }
        return edges;
    }

    /**
     * compute and return the list of triangles
     * 
     * @return 
     */
    public List<ValuedPoint2D[]> computeTriangles() {
        List<ValuedPoint2D[]> triangles = new ArrayList<>();

        // do not process edges pointing to/from surrouding triangle
        // --> mark them as already computed
        for (QuadEdge q : this.quadEdge) {
            q.mark = false;
            q.sym().mark = false;
            if (q.orig() == bbox.a || q.orig() == bbox.b || q.orig() == bbox.c || q.orig() == bbox.d) {
                q.mark = true;
            }
            if (q.dest() == bbox.a || q.dest() == bbox.b || q.dest() == bbox.c || q.dest() == bbox.d) {
                q.sym().mark = true;
            }
        }

        // compute the 2 triangles associated to each quadEdge
        for (QuadEdge qe : quadEdge) {
            // first triangle
            QuadEdge q1 = qe;
            QuadEdge q2 = q1.lnext();
            QuadEdge q3 = q2.lnext();
            if (!q1.mark && !q2.mark && !q3.mark) {
                triangles.add(new ValuedPoint2D[]{q1.orig(), q2.orig(), q3.orig()});
            }

            // second triangle
            QuadEdge qsym1 = qe.sym();
            QuadEdge qsym2 = qsym1.lnext();
            QuadEdge qsym3 = qsym2.lnext();
            if (!qsym1.mark && !qsym2.mark && !qsym3.mark) {
                triangles.add(new ValuedPoint2D[]{qsym1.orig(), qsym2.orig(), qsym3.orig()});
            }

            // mark as used
            qe.mark = true;
            qe.sym().mark = true;
        }

        return triangles;
    }

    public List<ValuedPoint2D[]> computeVoronoi() {
        List<ValuedPoint2D[]> voronoi = new ArrayList<>();

        // do not process edges pointing to/from surrouding triangle
        // --> mark them as already computed
        for (QuadEdge q : this.quadEdge) {
            q.mark = false;
            q.sym().mark = false;
            if (q.orig() == bbox.a || q.orig() == bbox.b || q.orig() == bbox.c || q.orig() == bbox.d) {
                q.mark = true;
            }
            if (q.dest() == bbox.a || q.dest() == bbox.b || q.dest() == bbox.c || q.dest() == bbox.d) {
                q.sym().mark = true;
            }
        }

        for (QuadEdge qe : quadEdge) {

            // walk throught left and right region
            for (int b = 0; b <= 1; b++) {
                QuadEdge qstart = (b == 0) ? qe : qe.sym();
                if (qstart.mark) {
                    continue;
                }

                // new region start
                List<ValuedPoint2D> poly = new ArrayList<>();
                poly.add(qstart.orig());

                // walk around region
                QuadEdge qregion = qstart;
                while (true) {
                    qregion.mark = true;

                    // compute CircumCenter if needed
                    if (qregion.rot().orig() == null) {
                        QuadEdge q1 = qregion;
                        ValuedPoint2D p0 = q1.orig();
                        QuadEdge q2 = q1.lnext();
                        ValuedPoint2D p1 = q2.orig();
                        QuadEdge q3 = q2.lnext();
                        ValuedPoint2D p2 = q3.orig();

                        double ex = p1.getX() - p0.getX(), ey = p1.getY() - p0.getY();
                        double nx = p2.getY() - p1.getY(), ny = p1.getX() - p2.getX();
                        double dx = (p0.getX() - p2.getX()) * 0.5, dy = (p0.getY() - p2.getY()) * 0.5;
                        double s = (ex * dx + ey * dy) / (ex * nx + ey * ny);
                        double cx = (p1.getX() + p2.getX()) * 0.5 + s * nx;
                        double cy = (p1.getY() + p2.getY()) * 0.5 + s * ny;

                        java.awt.geom.Point2D p = new java.awt.geom.Point2D.Double(cx, cy);
                        qregion.rot().setOrig(new ValuedPoint2D(p, null));
                    }

                    poly.add(qregion.rot().orig());

                    qregion = qregion.onext();
                    if (qregion == qstart) {
                        break;
                    }
                }

                // add region to output list
                voronoi.add(poly.toArray(new ValuedPoint2D[0]));
            }
        }
        return voronoi;
    }
}

/**
 * Quad-Edge data structure
 *
 * @author Code by X.Philippeau - Structure by Guibas and Stolfi
 *
 * @see Primitives for the Manipulation of General Subdivisions and the
 * Computation of Voronoi Diagrams (Leonidas Guibas,Jorge Stolfi)
 */
class QuadEdge {

    // pointer to the next (direct order) QuadEdge
    private QuadEdge onext;
    // pointer to the dual QuadEdge (faces graph <-> edges graph)
    private QuadEdge rot;
    // origin point of the edge/face
    private ValuedPoint2D orig;
    // marker for triangle generation
    public boolean mark = false;

    /**
     * (private) constructor. Use makeEdge() to create a new QuadEdge
     *
     * @param onext pointer to the next QuadEdge on the ring
     * @param rot pointer to the next (direct order) crossing edge
     * @param orig Origin point
     */
    private QuadEdge(QuadEdge Onext, QuadEdge rot, ValuedPoint2D orig) {
        this.onext = Onext;
        this.rot = rot;
        this.orig = orig;
    }

    // ----------------------------------------------------------------
    //                             Getter/Setter
    // ----------------------------------------------------------------
    public QuadEdge onext() {
        return onext;
    }

    public QuadEdge rot() {
        return rot;
    }

    public ValuedPoint2D orig() {
        return orig;
    }

    public void setOnext(QuadEdge next) {
        onext = next;
    }

    public void setRot(QuadEdge rot) {
        this.rot = rot;
    }

    public void setOrig(ValuedPoint2D p) {
        this.orig = p;
    }

    // ----------------------------------------------------------------
    //                      QuadEdge Navigation
    // ----------------------------------------------------------------
    /**
     * @return the symetric (reverse) QuadEdge
     */
    public QuadEdge sym() {
        return rot.rot();
    }

    /**
     * @return the other extremity point
     */
    public ValuedPoint2D dest() {
        return sym().orig();
    }

    /**
     * @return the symetric dual QuadEdge
     */
    public QuadEdge rotSym() {
        return rot.sym();
    }

    /**
     * @return the previous QuadEdge (pointing to this.orig)
     */
    public QuadEdge oprev() {
        return rot.onext().rot();
    }

    /**
     * @return the previous QuadEdge starting from dest()
     */
    public QuadEdge dprev() {
        return rotSym().onext().rotSym();
    }

    /**
     * @return the next QuadEdge on left Face
     */
    public QuadEdge lnext() {
        return rotSym().onext().rot();
    }

    /**
     * @return the previous QuadEdge on left Face
     */
    public QuadEdge lprev() {
        return onext().sym();
    }

    // ************************** STATIC ******************************
    /**
     * Create a new edge (i.e. a segment)
     *
     * @param orig origin of the segment
     * @param dest end of the segment
     * @return the QuadEdge of the origin point
     */
    public static QuadEdge makeEdge(ValuedPoint2D orig, ValuedPoint2D dest) {
        QuadEdge q0 = new QuadEdge(null, null, orig);
        QuadEdge q1 = new QuadEdge(null, null, null);
        QuadEdge q2 = new QuadEdge(null, null, dest);
        QuadEdge q3 = new QuadEdge(null, null, null);

        // create the segment
        q0.onext = q0;
        q2.onext = q2; // lonely segment: no "next" quadedge
        q1.onext = q3;
        q3.onext = q1; // in the dual: 2 communicating facets

        // dual switch
        q0.rot = q1;
        q1.rot = q2;
        q2.rot = q3;
        q3.rot = q0;

        return q0;
    }

    /**
     * attach/detach the two edges = combine/split the two rings in the dual
     * space
     *
     * @param q1,q2 the 2 QuadEdge to attach/detach
     */
    public static void splice(QuadEdge a, QuadEdge b) {
        QuadEdge alpha = a.onext().rot();
        QuadEdge beta = b.onext().rot();

        QuadEdge t1 = b.onext();
        QuadEdge t2 = a.onext();
        QuadEdge t3 = beta.onext();
        QuadEdge t4 = alpha.onext();

        a.setOnext(t1);
        b.setOnext(t2);
        alpha.setOnext(t3);
        beta.setOnext(t4);
    }

    /**
     * Create a new QuadEdge by connecting 2 QuadEdges
     *
     * @param e1,e2 the 2 QuadEdges to connect
     * @return the new QuadEdge
     */
    public static QuadEdge connect(QuadEdge e1, QuadEdge e2) {
        QuadEdge q = makeEdge(e1.dest(), e2.orig());
        splice(q, e1.lnext());
        splice(q.sym(), e2);
        return q;
    }

    public static void swapEdge(QuadEdge e) {
        QuadEdge a = e.oprev();
        QuadEdge b = e.sym().oprev();
        splice(e, a);
        splice(e.sym(), b);
        splice(e, a.lnext());
        splice(e.sym(), b.lnext());
        e.orig = a.dest();
        e.sym().orig = b.dest();
    }

    /**
     * Delete a QuadEdge
     *
     * @param q the QuadEdge to delete
     */
    public static void deleteEdge(QuadEdge q) {
        splice(q, q.oprev());
        splice(q.sym(), q.sym().oprev());
    }

    // ----------------------------------------------------------------
    //                      Geometric computation
    // ----------------------------------------------------------------
    /**
     * Test if the ValuedPoint2D p is on the edge
     *
     * @param e QuadEdge
     * @param p ValuedPoint2D to test
     * @return true/false
     */
    public static boolean isOnEdge(QuadEdge e, ValuedPoint2D p) {
        double d = distanceToSegment(e.orig(), e.dest(), p);
        return d <= 0.01;
    }

    /**
     * Return the distance from a point to a segment
     *
     * @param ps,pe the start/end of the segment
     * @param p the
     * @return
     */
    private static double distanceToSegment(ValuedPoint2D ps, ValuedPoint2D pe, ValuedPoint2D p) {
        if (ps.getX() == pe.getX() && ps.getY() == pe.getY()) {
            return squaredistance(ps, p);
        }

        double sx = pe.getX() - ps.getX();
        double sy = pe.getY() - ps.getY();

        double ux = p.getX() - ps.getX();
        double uy = p.getY() - ps.getY();

        double dp = sx * ux + sy * uy;
        if (dp < 0) {
            return squaredistance(ps, p);
        }

        double sn = sx * sx + sy * sy;
        if (dp > sn) {
            return squaredistance(pe, p);
        }

        double ratio = dp / sn;
        int projx = (int) (ps.getX() + ratio * sx);
        int projy = (int) (ps.getY() + ratio * sy);

        return squaredistance(new ValuedPoint2D(projx, projy), p);
    }

    /**
     * return the square of the distance between two points
     *
     * @param p1,p2 the two points
     * @return dist the square of the distance
     */
    private static double squaredistance(ValuedPoint2D p1, ValuedPoint2D p2) {
        double d2 = (p2.getX() - p1.getX()) * (p2.getX() - p1.getX()) + (p2.getY() - p1.getY()) * (p2.getY() - p1.getY());
        return d2;
    }

    /**
     * Test if the ValuedPoint2D p is at the right of the QuadEdge q.
     *
     * @param QuadEdge reference
     * @param p ValuedPoint2D to test
     * @return true/false
     */
    public static boolean isAtRightOf(QuadEdge q, ValuedPoint2D p) {
        return isCounterClockwise(p, q.dest(), q.orig());
    }

    /**
     * return true if a, b and c turn in Counter Clockwise direction
     *
     * @param a,b,c the 3 points to test
     * @return true if a, b and c turn in Counter Clockwise direction
     */
    public static boolean isCounterClockwise(ValuedPoint2D a, ValuedPoint2D b, ValuedPoint2D c) {
        return (a.getX() - b.getX()) * (b.getY() - c.getY()) > (a.getY() - b.getY()) * (b.getX() - c.getX());
    }

    /**
     * The Delaunay criteria:
     *
     * test if the point d is inside the circumscribed circle of triangle a,b,c
     *
     * @param a,b,c triangle
     * @param d point to test
     * @return true/false
     */
    public static boolean inCircle(ValuedPoint2D a, ValuedPoint2D b, ValuedPoint2D c, ValuedPoint2D d) {
        /**
         * if "d" is strictly INSIDE the circle, the
         */
        double a2 = a.getX() * a.getX() + a.getY() * a.getY();
        double b2 = b.getX() * b.getX() + b.getY() * b.getY();
        double c2 = c.getX() * c.getX() + c.getY() * c.getY();
        double d2 = d.getX() * d.getX() + d.getY() * d.getY();

        double det44 = 0;
        det44 += d2 * det33(a.getX(), a.getY(), 1, b.getX(), b.getY(), 1, c.getX(), c.getY(), 1);
        det44 -= d.getX() * det33(a2, a.getY(), 1, b2, b.getY(), 1, c2, c.getY(), 1);
        det44 += d.getY() * det33(a2, a.getX(), 1, b2, b.getX(), 1, c2, c.getX(), 1);
        det44 -= 1 * det33(a2, a.getX(), a.getY(), b2, b.getX(), b.getY(), c2, c.getX(), c.getY());
        return det44 < 0;
    }

    private static double det33(double... m) {
        double det33 = 0;
        det33 += m[0] * (m[4] * m[8] - m[5] * m[7]);
        det33 -= m[1] * (m[3] * m[8] - m[5] * m[6]);
        det33 += m[2] * (m[3] * m[7] - m[4] * m[6]);
        return det33;
    }
}